from abc import ABCMeta, abstractmethod, abstractproperty
from functools import lru_cache

from atlassian_jwt_auth import HTTPSPublicKeyRetriever, JWTAuthVerifier

from .utils import SettingsDict


@lru_cache(maxsize=20)
def _get_verifier(settings):
    """ This has been extracted out of Backend to avoid possible memory
        leaks via retained instance references.
    """
    retriever = settings.ASAP_KEY_RETRIEVER_CLASS(
        base_url=settings.ASAP_PUBLICKEY_REPOSITORY
    )
    kwargs = {}
    if settings.ASAP_SUBJECT_SHOULD_MATCH_ISSUER is not None:
        kwargs = {'subject_should_match_issuer':
                  settings.ASAP_SUBJECT_SHOULD_MATCH_ISSUER}
    if settings.ASAP_CHECK_JTI_UNIQUENESS is not None:
        kwargs['check_jti_uniqueness'] = settings.ASAP_CHECK_JTI_UNIQUENESS
    return JWTAuthVerifier(
        retriever,
        **kwargs
    )


class Backend():
    """Abstract class representing a web framework backend

    Backends allow specific implementation details of web frameworks to be
    abstracted away from the underlying logic of ASAP.
    """
    __metaclass__ = ABCMeta

    default_headers_401 = {'WWW-Authenticate': 'Bearer'}
    default_settings = {
        # The class to be instantiated to retrieve public keys
        'ASAP_KEY_RETRIEVER_CLASS': HTTPSPublicKeyRetriever,

        # The repository URL where the key retriever can fetch public keys
        'ASAP_PUBLICKEY_REPOSITORY': None,

        # Whether or not ASAP authentication is required
        # This is primarily useful when phasing in ASAP authentication
        'ASAP_REQUIRED': True,

        # The valid audience value expected when authenticating tokens
        'ASAP_VALID_AUDIENCE': None,

        # The amount of leeway to apply when evaluating token expiration
        # timestamps
        'ASAP_VALID_LEEWAY': 0,

        # An iterable of valid token issuers allowed to authenticate
        # (this can be overridden at the decorator level)
        'ASAP_VALID_ISSUERS': None,

        # Enforce that the ASAP subject must match the issuer
        'ASAP_SUBJECT_SHOULD_MATCH_ISSUER': None,

        # Enforce that tokens have a unique JTI
        # Set this to True to enforce JTI uniqueness checking.
        'ASAP_CHECK_JTI_UNIQUENESS': None,
    }

    @abstractmethod
    def get_authorization_header(self, request=None):
        pass

    @abstractmethod
    def get_401_response(self, data=None, headers=None, request=None):
        pass

    @abstractmethod
    def get_403_response(self, data=None, headers=None, request=None):
        pass

    @abstractmethod
    def set_asap_claims_for_request(self, request, claims):
        pass

    @abstractproperty
    def settings(self):
        return SettingsDict(self.default_settings)

    def get_asap_token(self, request):
        auth_header = self.get_authorization_header(request)

        if auth_header is None:
            return None

        if isinstance(auth_header, str):
            # Per PEP-3333, headers must be in ISO-8859-1 or use an RFC-2047
            # MIME encoding. We don't really care about MIME encoded
            # headers, but some libraries allow sending bytes (Django tests)
            # and some (requests) always send str so we need to convert if
            # that is the case to properly support Python 3.
            auth_header = auth_header.encode(encoding='iso-8859-1')

        auth_values = auth_header.split(b' ')
        if len(auth_values) != 2 or auth_values[0].lower() != b'bearer':
            return None

        return auth_values[1]

    def get_verifier(self, settings=None):
        """Returns a verifier for ASAP JWT tokens"""
        if settings is None:
            settings = self.settings
        return self._get_verifier(settings)

    def _get_verifier(self, settings):
        return _get_verifier(settings)

    def _process_settings(self, settings):
        valid_issuers = settings.get('ASAP_VALID_ISSUERS')
        if valid_issuers:
            settings['ASAP_VALID_ISSUERS'] = set(valid_issuers)

        return SettingsDict(settings)



import os
import re
import unittest
from unittest import mock

import httptest
import requests

from atlassian_jwt_auth.key import (
    HTTPSPublicKeyRetriever,
    HTTPSMultiRepositoryPublicKeyRetriever,
    PEM_FILE_TYPE,
)
from atlassian_jwt_auth.tests import utils


def get_expected_and_os_proxies_dict(proxy_location):
    """ returns expected proxy & environmental
        proxy dictionary based upon the provided proxy location.
    """
    expected_proxies = {
        'http': proxy_location,
        'https': proxy_location,
    }
    os_proxy_dict = {
        'HTTP_PROXY': proxy_location,
        'HTTPS_PROXY': proxy_location
    }
    return expected_proxies, os_proxy_dict


class BaseHTTPSPublicKeyRetrieverTest(object):
    """ tests for the HTTPSPublicKeyRetriever class. """

    def create_retriever(self, url):
        """ returns a public key retriever created using the given url. """
        return HTTPSPublicKeyRetriever(url)

    def setUp(self):
        self._private_key_pem = self.get_new_private_key_in_pem_format()
        self._public_key_pem = utils.get_public_key_pem_for_private_key_pem(
            self._private_key_pem)
        self.base_url = 'https://example.com'

    def test_https_public_key_retriever_does_not_support_http_url(self):
        """ tests that HTTPSPublicKeyRetriever does not support http://
            base urls.
        """
        with self.assertRaises(ValueError):
            self.create_retriever('http://example.com')

    def test_https_public_key_retriever_does_not_support_none_url(self):
        """ tests that HTTPSPublicKeyRetriever does not support None
            base urls.
        """
        with self.assertRaises(ValueError):
            self.create_retriever(None)

    def test_https_public_key_retriever_session_uses_env_proxy(self):
        """ tests that the underlying session makes use of environmental
            proxy configured.
        """
        proxy_location = 'https://example.proxy'
        expected_proxies, proxy_dict = get_expected_and_os_proxies_dict(
            proxy_location)
        with mock.patch.dict(os.environ, proxy_dict, clear=True):
            retriever = self.create_retriever(self.base_url)
            key_retrievers = [retriever]
            if isinstance(retriever, HTTPSMultiRepositoryPublicKeyRetriever):
                key_retrievers = retriever._retrievers
            for key_retriever in key_retrievers:
                self.assertEqual(key_retriever._proxies, expected_proxies)

    def test_https_public_key_retriever_supports_https_url(self):
        """ tests that HTTPSPublicKeyRetriever supports https://
            base urls.
        """
        self.create_retriever(self.base_url)

    @mock.patch.object(requests.Session, 'get')
    def test_retrieve(self, mock_get_method):
        """ tests that the retrieve method works expected. """
        _setup_mock_response_for_retriever(
            mock_get_method, self._public_key_pem)
        retriever = self.create_retriever(self.base_url)
        self.assertEqual(
            retriever.retrieve('example/eg'),
            self._public_key_pem)

    @mock.patch.object(requests.Session, 'get')
    def test_retrieve_with_proxy(self, mock_get_method):
        """ tests that the retrieve method works as expected when a proxy
            should be used.
        """
        proxy_location = 'https://example.proxy'
        key_id = 'example/eg'
        expected_proxies, proxy_dict = get_expected_and_os_proxies_dict(
            proxy_location)
        _setup_mock_response_for_retriever(
            mock_get_method, self._public_key_pem)
        with mock.patch.dict(os.environ, proxy_dict, clear=True):
            retriever = self.create_retriever(self.base_url)
            retriever.retrieve(key_id)
            mock_get_method.assert_called_once_with(
                '%s/%s' % (self.base_url, key_id),
                headers={'accept': PEM_FILE_TYPE},
                proxies=expected_proxies
            )

    @mock.patch.object(requests.Session, 'get')
    def test_retrieve_with_proxy_explicitly_set(self, mock_get_method):
        """ tests that the retrieve method works as expected when a proxy
            should be used and has been explicitly provided.
        """
        proxy_location = 'https://example.proxy'
        explicit_proxy_location = 'https://explicit.proxy'
        key_id = 'example/eg'
        _, proxy_dict = get_expected_and_os_proxies_dict(proxy_location)
        expected_proxies, _ = get_expected_and_os_proxies_dict(
            explicit_proxy_location)
        _setup_mock_response_for_retriever(
            mock_get_method, self._public_key_pem)
        with mock.patch.dict(os.environ, proxy_dict, clear=True):
            retriever = self.create_retriever(self.base_url)
            retriever.retrieve(key_id, proxies=expected_proxies)
            mock_get_method.assert_called_once_with(
                '%s/%s' % (self.base_url, key_id),
                headers={'accept': PEM_FILE_TYPE},
                proxies=expected_proxies
            )

    @mock.patch.object(requests.Session, 'get')
    def test_retrieve_with_charset_in_content_type_h(self, mock_get_method):
        """ tests that the retrieve method works expected when there is
            a charset in the response content-type header.
        """
        headers = {'content-type': 'application/x-pem-file;charset=UTF-8'}
        _setup_mock_response_for_retriever(
            mock_get_method, self._public_key_pem, headers)
        retriever = self.create_retriever(self.base_url)
        self.assertEqual(
            retriever.retrieve('example/eg'),
            self._public_key_pem)

    @mock.patch.object(requests.Session, 'get')
    def test_retrieve_fails_with_different_content_type(self, mock_get_method):
        """ tests that the retrieve method fails when the response is for a
            media type that is not supported.
        """
        headers = {'content-type': 'different/not-supported'}
        _setup_mock_response_for_retriever(
            mock_get_method, self._public_key_pem, headers)
        retriever = self.create_retriever(self.base_url)
        with self.assertRaises(ValueError):
            retriever.retrieve('example/eg')

    @mock.patch.object(requests.Session, 'get',
                       side_effect=requests.exceptions.HTTPError(
                           mock.Mock(response=mock.Mock(status_code=403)),
                           'forbidden'))
    def test_retrieve_fails_with_forbidden_error(self, mock_get_method):
        """ tests that the retrieve method fails when the response is an
        403 forbidden error.
        """
        _setup_mock_response_for_retriever(
            mock_get_method, self._public_key_pem)
        retriever = self.create_retriever(self.base_url)
        with self.assertRaises(ValueError):
            retriever.retrieve('example/eg')


class CachedHTTPPublicKeyRetrieverTest(utils.ES256KeyTestMixin,
                                       unittest.TestCase):

    class HTTPPublicKeyRetriever(HTTPSPublicKeyRetriever):
        """A subclass of HTTPSPublicKeyRetriever that allows us to use plain
        HTTP during testing so we don't have to run an actual SSL server.
        """

        def __init__(self, base_url):
            # pretend to the super class that this is an HTTPS url
            super(CachedHTTPPublicKeyRetrieverTest.HTTPPublicKeyRetriever,
                  self).__init__(
                re.sub(r'^http', 'https', base_url, flags=re.IGNORECASE))
            self.base_url = base_url

    def setUp(self):
        super(CachedHTTPPublicKeyRetrieverTest, self).setUp()
        self._private_key_pem = self.get_new_private_key_in_pem_format()
        self._public_key_pem = utils.get_public_key_pem_for_private_key_pem(
            self._private_key_pem)

    def test_http_caching(self):
        """Asserts that our use of requests properly caches keys between
        invocations across different `HTTPSPublicKeyRetriever` instances.
        """
        def wsgi(environ, start_response):
            print(environ['PATH_INFO'])
            start_response('200 OK', [
                ('content-type', 'application/x-pem-file;charset=UTF-8'),
                ('Cache-Control', 'public,max-age=300,stale-while-revalidate='
                                  '300,stale-if-error=300'),
                ('Last-Modified', 'Sun, 18 Jan 1970 18:14:21 GMT')])
            return [self._public_key_pem]

        with httptest.testserver(wsgi) as server:

            retriever = self.HTTPPublicKeyRetriever(server.url())
            retriever.retrieve('example/eg')

            retriever = self.HTTPPublicKeyRetriever(server.url())
            retriever.retrieve('example/eg')

            self.assertEqual(1, len(server.log()),
                             msg='HTTP caching should suppress second GET')


class BaseHTTPSMultiRepositoryPublicKeyRetrieverTest(
        BaseHTTPSPublicKeyRetrieverTest):
    """ tests for the HTTPSMultiRepositoryPublicKeyRetriever class. """

    def create_retriever(self, url):
        """ returns a public key retriever created using the given url. """
        return HTTPSMultiRepositoryPublicKeyRetriever([url])

    def setUp(self):
        self._private_key_pem = self.get_new_private_key_in_pem_format()
        self._public_key_pem = utils.get_public_key_pem_for_private_key_pem(
            self._private_key_pem)
        self.keystore_urls = ['https://example.com', 'https://example.ly']
        self.base_url = self.keystore_urls[0]

    def test_https_multi_public_key_retriever_does_not_support_strings(self):
        """ tests that HTTPSMultiRepositoryPublicKeyRetriever does not
            support a string key repository url.
        """
        with self.assertRaises(TypeError):
            HTTPSMultiRepositoryPublicKeyRetriever('https://example.com')

    @mock.patch.object(requests.Session, 'get')
    def test_retrieve(self, mock_get_method):
        """ tests that the retrieve method works expected. """
        _setup_mock_response_for_retriever(
            mock_get_method, self._public_key_pem)
        retriever = HTTPSMultiRepositoryPublicKeyRetriever(self.keystore_urls)
        self.assertEqual(
            retriever.retrieve('example/eg'),
            self._public_key_pem)

    @mock.patch.object(requests.Session, 'get')
    def test_retrieve_with_500_error(self, mock_get_method):
        """ tests that the retrieve method works as expected
            when the first key repository returns a server error response.
        """
        retriever = HTTPSMultiRepositoryPublicKeyRetriever(self.keystore_urls)
        _setup_mock_response_for_retriever(
            mock_get_method, self._public_key_pem)
        valid_response = mock_get_method.return_value
        del mock_get_method.return_value
        server_exception = requests.exceptions.HTTPError(
            response=mock.Mock(status_code=500))
        mock_get_method.side_effect = [server_exception, valid_response]
        self.assertEqual(
            retriever.retrieve('example/eg'),
            self._public_key_pem)

    @mock.patch.object(requests.Session, 'get')
    def test_retrieve_with_connection_error(self, mock_get_method):
        """ tests that the retrieve method works as expected
            when the first key repository encounters a connection error.
        """
        retriever = HTTPSMultiRepositoryPublicKeyRetriever(self.keystore_urls)
        _setup_mock_response_for_retriever(
            mock_get_method, self._public_key_pem)
        valid_response = mock_get_method.return_value
        del mock_get_method.return_value
        connection_exception = requests.exceptions.ConnectionError(
            response=mock.Mock(status_code=None))
        mock_get_method.side_effect = [connection_exception, valid_response]
        self.assertEqual(
            retriever.retrieve('example/eg'),
            self._public_key_pem)


def _setup_mock_response_for_retriever(
        mock_method, public_key_pem, headers=None):
    """ returns a setup mock response for use with a https public key
        retriever.
    """
    if headers is None:
        headers = {'content-type': 'application/x-pem-file'}
    mock_response = mock.Mock()
    mock_response.headers = headers
    mock_response.text = public_key_pem
    mock_method.return_value = mock_response
    return mock_method


class HTTPSPublicKeyRetrieverRS256Test(BaseHTTPSPublicKeyRetrieverTest,
                                       utils.RS256KeyTestMixin,
                                       unittest.TestCase):
    pass


class HTTPSPublicKeyRetrieverES256Test(BaseHTTPSPublicKeyRetrieverTest,
                                       utils.ES256KeyTestMixin,
                                       unittest.TestCase):
    pass


class HTTPSMultiRepositoryPublicKeyRetrieverRS256Test(
        BaseHTTPSMultiRepositoryPublicKeyRetrieverTest,
        utils.RS256KeyTestMixin,
        unittest.TestCase):
    pass


class HTTPSMultiRepositoryPublicKeyRetrieverES256Test(
        BaseHTTPSMultiRepositoryPublicKeyRetrieverTest,
        utils.ES256KeyTestMixin,
        unittest.TestCase):
    pass



import asyncio

import jwt

from atlassian_jwt_auth import key
from atlassian_jwt_auth.verifier import JWTAuthVerifier as _JWTAuthVerifier


class JWTAuthVerifier(_JWTAuthVerifier):
    async def verify_jwt(self, a_jwt, audience, leeway=0, **requests_kwargs):
        """Verify if the token is correct

        Returns:
             dict: the claims of the given jwt if verification is successful.

        Raises:
            ValueError: if verification failed.
        """
        key_identifier = key._get_key_id_from_jwt_header(a_jwt)

        public_key = self._retrieve_pub_key(key_identifier, requests_kwargs)
        if asyncio.iscoroutine(public_key):
            public_key = await public_key

        alg = jwt.get_unverified_header(a_jwt).get('alg', None)
        public_key_obj = self._load_public_key(public_key, alg)
        return self._decode_jwt(
            a_jwt, key_identifier, public_key_obj,
            audience=audience, leeway=leeway)



from ..common.backend import Backend
from ..common.utils import SettingsDict


class WSGIBackend(Backend):
    def __init__(self, settings):
        self._settings = SettingsDict(settings)

    def get_authorization_header(self, request=None):
        if request is None:
            raise ValueError('No request available')

        return request.environ.get('HTTP_AUTHORIZATION', b'')

    def get_401_response(self, data=None, headers=None, request=None):
        if request is None:
            raise TypeError("request must have a value")

        if headers is None:
            headers = {}

        headers.update(self.default_headers_401)

        request.start_response('401 Unauthorized', list(headers.items()), None)
        return ""

    def get_403_response(self, data=None, headers=None, request=None):
        if request is None:
            raise TypeError("request must have a value")

        if headers is None:
            headers = {}

        request.start_response('403 Forbidden', list(headers.items()), None)
        return ""

    def set_asap_claims_for_request(self, request, claims):
        request.environ['ATL_ASAP_CLAIMS'] = claims

    @property
    def settings(self):
        settings = {}
        settings.update(self.default_settings)

        for k in settings.keys():
            value = getattr(self._settings, k, None)
            if value is None:
                continue

            settings[k] = value

        return self._process_settings(settings)



from django.urls import path

from atlassian_jwt_auth.frameworks.django.tests import views


urlpatterns = [
    path('asap/expected', views.expected_view, name='expected'),
    path(r'^asap/unexpected', views.unexpected_view, name='unexpected'),
    path('^asap/decorated', views.decorated_view, name='decorated'),
    path('asap/settings', views.settings_view, name='settings'),

    path('asap/subject_does_not_need_to_match_issuer',
         views.subject_does_not_need_to_match_issuer_view,
         name='subject_does_not_need_to_match_issuer'),
    path('asap/subject_does_need_to_match_issuer_view',
         views.subject_does_need_to_match_issuer_view,
         name='subject_does_need_to_match_issuer'),

    path('asap/subject_does_not_need_to_match_issuer_from_settings',
         views.subject_does_not_need_to_match_issuer_from_settings_view,
         name='subject_does_not_need_to_match_issuer_from_settings'),

    path('asap/needed', views.needed_view, name='needed'),
    path(r'asap/unneeded', views.unneeded_view, name='unneeded'),
    path(r'asap/restricted_issuer', views.restricted_issuer_view,
         name='restricted_issuer'),
    path('asap/restricted_subject', views.restricted_subject_view,
         name='restricted_subject'),
]



from django.http import HttpResponse

from atlassian_jwt_auth.frameworks.django import with_asap, restrict_asap
from atlassian_jwt_auth.contrib.django.decorators import (requires_asap,
                                                          validate_asap)


@with_asap(issuers=['client-app'])
def expected_view(request):
    return HttpResponse('Greatest Success!')


@with_asap(issuers=['unexpected'])
def unexpected_view(request):
    return HttpResponse('This should fail.')


@with_asap(issuers=['whitelist'])
def decorated_view(request):
    return HttpResponse('Only the right issuer is allowed.')


@requires_asap()
def settings_view(request):
    return HttpResponse('Any settings issuer is allowed.')


@with_asap(subject_should_match_issuer=False)
def subject_does_not_need_to_match_issuer_view(request):
    return HttpResponse('Subject does not need to match issuer.')


@with_asap(subject_should_match_issuer=True)
def subject_does_need_to_match_issuer_view(request):
    return HttpResponse('Subject does need to match issuer.')


@with_asap()
def subject_does_not_need_to_match_issuer_from_settings_view(request):
    return HttpResponse('Subject does not need to match issuer (settings).')


@restrict_asap
def needed_view(request):
    return HttpResponse('one')


@restrict_asap(required=False)
def unneeded_view(request):
    return HttpResponse('two')


@restrict_asap(issuers=['whitelist'])
def restricted_issuer_view(request):
    return HttpResponse('three')


@validate_asap(subjects=['client-app'])
def restricted_subject_view(request):
    return HttpResponse('four')



import os
from unittest import mock

import aiohttp
from multidict import CIMultiDict

try:
    from unittest import IsolatedAsyncioTestCase as TestCase
    from unittest.mock import AsyncMock as CoroutineMock
    from unittest.mock import Mock
except ImportError:
    from asynctest import CoroutineMock, TestCase, Mock

from atlassian_jwt_auth.contrib.aiohttp import HTTPSPublicKeyRetriever
from atlassian_jwt_auth.key import PEM_FILE_TYPE
from atlassian_jwt_auth.tests import utils
from atlassian_jwt_auth.tests.test_public_key_provider import (
    get_expected_and_os_proxies_dict,
)


class DummyHTTPSPublicKeyRetriever(HTTPSPublicKeyRetriever):

    def set_headers(self, headers):
        self._session.get.return_value.headers.update(headers)

    def set_text(self, text):
        self._session.get.return_value.text.return_value = text

    def _get_session(self):
        session = Mock(spec=aiohttp.ClientSession)
        session.attach_mock(CoroutineMock(), 'get')

        resp = session.get.return_value
        resp.headers = CIMultiDict({"content-type": PEM_FILE_TYPE})
        resp.text = CoroutineMock(return_value='i-am-a-public-key')
        resp.raise_for_status = Mock(name='raise_for_status')
        return session


class BaseHTTPSPublicKeyRetrieverTestMixin(object):
    """Tests for aiohttp.HTTPSPublicKeyRetriever class for RS256 algorithm"""

    def setUp(self):
        self._private_key_pem = self.get_new_private_key_in_pem_format()
        self._public_key_pem = utils.get_public_key_pem_for_private_key_pem(
            self._private_key_pem)
        self.base_url = 'https://example.com'

    async def test_retrieve(self):
        """Check if retrieve method returns public key"""
        retriever = DummyHTTPSPublicKeyRetriever(self.base_url)
        retriever.set_text(self._public_key_pem)
        self.assertEqual(
            await retriever.retrieve('example/eg'),
            self._public_key_pem)

    async def test_retrieve_with_charset_in_content_type_h(self):
        """Check if retrieve method correctly checks content-type"""
        headers = {'content-type': 'application/x-pem-file;charset=UTF-8'}
        retriever = DummyHTTPSPublicKeyRetriever(self.base_url)
        retriever.set_text(self._public_key_pem)
        retriever.set_headers(headers)

        self.assertEqual(
            await retriever.retrieve('example/eg'),
            self._public_key_pem)

    async def test_retrieve_fails_with_different_content_type(self):
        """
        Check if retrieve method raises an error for incorrect content-type
        """
        headers = {'content-type': 'different/not-supported'}
        retriever = DummyHTTPSPublicKeyRetriever(self.base_url)
        retriever.set_text(self._public_key_pem)
        retriever.set_headers(headers)

        with self.assertRaises(ValueError):
            await retriever.retrieve('example/eg')

    async def test_retrieve_session_uses_env_proxy(self):
        """ tests that the underlying session makes use of environmental
            proxy configured.
        """
        proxy_location = 'https://example.proxy'
        key_id = 'example/eg'
        expected_proxies, proxy_dict = get_expected_and_os_proxies_dict(
            proxy_location)
        with mock.patch.dict(os.environ, proxy_dict, clear=True):
            retriever = DummyHTTPSPublicKeyRetriever(self.base_url)
            self.assertEqual(retriever._proxies, expected_proxies)
            await retriever.retrieve(key_id)
            retriever._session.get.assert_called_once_with(
                f'{self.base_url}/{key_id}', headers={'accept': PEM_FILE_TYPE},
                proxy=expected_proxies[self.base_url.split(':')[0]]
            )


class RS256HTTPSPublicKeyRetrieverTest(utils.RS256KeyTestMixin,
                                       BaseHTTPSPublicKeyRetrieverTestMixin,
                                       TestCase):
    """Tests for aiohttp.HTTPSPublicKeyRetriever class for RS256 algorithm"""


class ES256HTTPSPublicKeyRetrieverTest(utils.RS256KeyTestMixin,
                                       BaseHTTPSPublicKeyRetrieverTestMixin,
                                       TestCase):
    """Tests for aiohttp.HTTPSPublicKeyRetriever class for ES256 algorithm"""



import unittest

from atlassian_jwt_auth.frameworks.common import utils


class SettingsDictTest(unittest.TestCase):
    """ Tests for the SettingsDict class. """

    def test_hash(self):
        """ Test that SettingsDict instances can be hashed. """
        dictionary_one = {'a': 'b', '3': set([1]), 'f': None}
        dictionary_two = {'a': 'b', '3': set([1]), 'f': None}
        dictionary_three = {'a': 'b', '3': set([1]), 'diff': '333'}
        settings_one = utils.SettingsDict(dictionary_one)
        settings_two = utils.SettingsDict(dictionary_two)
        settings_three = utils.SettingsDict(dictionary_three)
        self.assertEqual(settings_one, settings_two)
        self.assertEqual(hash(settings_one), hash(settings_two))
        self.assertNotEqual(settings_one, settings_three)
        self.assertNotEqual(hash(settings_one), hash(settings_three))



import unittest

from flask import Flask

from atlassian_jwt_auth.contrib.flask_app import requires_asap
from atlassian_jwt_auth.contrib.tests.utils import get_static_retriever_class
from atlassian_jwt_auth.frameworks.flask import with_asap
from atlassian_jwt_auth.tests import utils
from atlassian_jwt_auth.tests.utils import (
    create_token,
)


def get_app():
    app = Flask(__name__)
    app.config.update({
        'ASAP_VALID_AUDIENCE': 'server-app',
        'ASAP_VALID_ISSUERS': ('client-app',),
        'ASAP_PUBLICKEY_REPOSITORY': None
    })

    @app.route("/")
    @requires_asap
    def view():
        return "OK"

    @app.route("/restricted-to-another-client/")
    @with_asap(issuers=['another-client'])
    def view_for_another_client_app():
        return "OK"

    return app


class FlaskTests(utils.RS256KeyTestMixin, unittest.TestCase):
    """ tests for the atlassian_jwt_auth.contrib.tests.flask """

    def setUp(self):
        self._private_key_pem = self.get_new_private_key_in_pem_format()
        self._public_key_pem = utils.get_public_key_pem_for_private_key_pem(
            self._private_key_pem
        )

        self.app = get_app()
        self.client = self.app.test_client()

        retriever = get_static_retriever_class({
            'client-app/key01': self._public_key_pem
        })
        self.app.config['ASAP_KEY_RETRIEVER_CLASS'] = retriever

    def send_request(self, token, url='/'):
        """ returns the response of sending a request containing the given
            token sent in the Authorization header.
        """
        return self.client.get(url, headers={
            'Authorization': b'Bearer ' + token
        })

    def test_request_with_valid_token_is_allowed(self):
        token = create_token(
            'client-app', 'server-app',
            'client-app/key01', self._private_key_pem
        )
        self.assertEqual(self.send_request(token).status_code, 200)

    def test_request_with_duplicate_jti_is_rejected_as_per_setting(self):
        self.app.config['ASAP_CHECK_JTI_UNIQUENESS'] = True
        token = create_token(
            'client-app', 'server-app',
            'client-app/key01', self._private_key_pem
        )
        self.assertEqual(self.send_request(token).status_code, 200)
        self.assertEqual(self.send_request(token).status_code, 401)

    def _assert_request_with_duplicate_jti_is_accepted(self):
        token = create_token(
            'client-app', 'server-app',
            'client-app/key01', self._private_key_pem
        )
        self.assertEqual(self.send_request(token).status_code, 200)
        self.assertEqual(self.send_request(token).status_code, 200)

    def test_request_with_duplicate_jti_is_accepted(self):
        self._assert_request_with_duplicate_jti_is_accepted()

    def test_request_with_duplicate_jti_is_accepted_as_per_setting(self):
        self.app.config['ASAP_CHECK_JTI_UNIQUENESS'] = False
        self._assert_request_with_duplicate_jti_is_accepted()

    def test_request_with_invalid_audience_is_rejected(self):
        token = create_token(
            'client-app', 'invalid-audience',
            'client-app/key01', self._private_key_pem
        )
        self.assertEqual(self.send_request(token).status_code, 401)

    def test_request_with_invalid_token_is_rejected(self):
        response = self.send_request(b'notavalidtoken')
        self.assertEqual(response.status_code, 401)

    def test_request_with_invalid_issuer_is_rejected(self):
        # Try with a different audience with a valid signature
        self.app.config['ASAP_KEY_RETRIEVER_CLASS'] = (
            get_static_retriever_class({
                'another-client/key01': self._public_key_pem
            })
        )
        token = create_token(
            'another-client', 'server-app',
            'another-client/key01', self._private_key_pem
        )
        self.assertEqual(self.send_request(token).status_code, 403)

    def test_decorated_request_with_invalid_issuer_is_rejected(self):
        # Try with a different audience with a valid signature
        token = create_token(
            'client-app', 'server-app',
            'client-app/key01', self._private_key_pem
        )
        url = '/restricted-to-another-client/'
        self.assertEqual(self.send_request(token, url=url).status_code, 403)

    def test_request_subject_and_issue_not_matching(self):
        token = create_token(
            'client-app', 'server-app',
            'client-app/key01', self._private_key_pem,
            subject='different'
        )
        self.assertEqual(self.send_request(token).status_code, 401)

    def test_request_subject_does_not_need_to_match_issuer_from_settings(self):
        self.app.config['ASAP_SUBJECT_SHOULD_MATCH_ISSUER'] = False
        token = create_token(
            'client-app', 'server-app',
            'client-app/key01', self._private_key_pem,
            subject='different'
        )
        self.assertEqual(self.send_request(token).status_code, 200)



from ..common.decorators import _with_asap
from .backend import FlaskBackend


def with_asap(func=None, issuers=None, required=None,
              subject_should_match_issuer=None):
    """Decorator to allow endpoint-specific ASAP authentication.

    If authentication fails, a 401 or 403 response will be returned. Otherwise,
    the decorated function will be executed.

    The ASAP claimset will be set on g.asap_claims for further
    inspection later in the request lifecycle.

    :param list func: The view to decorate.
    :param list issuers: A list of valid token issuers that can access this
                         endpoint.
    :param boolean required: Whether or not to require ASAP on this endpoint.
    :param boolean subject_should_match_issuer: Indicate whether the subject
                                                must match the issuer for a
                                                token to be considered valid.
    """
    return _with_asap(
        func, FlaskBackend(), issuers, required,
        subject_should_match_issuer
    )



import unittest

from atlassian_jwt_auth.contrib.tests.utils import get_static_retriever_class
from atlassian_jwt_auth.frameworks.wsgi.middleware import ASAPMiddleware
from atlassian_jwt_auth.tests import utils
from atlassian_jwt_auth.tests.utils import (
    create_token,
)


def app(environ, start_response):
    start_response('200 OK', [], None)
    return "OK"


class WsgiTests(utils.RS256KeyTestMixin, unittest.TestCase):
    """ tests for the atlassian_jwt_auth.contrib.tests.flask """

    def setUp(self):
        self._private_key_pem = self.get_new_private_key_in_pem_format()
        self._public_key_pem = utils.get_public_key_pem_for_private_key_pem(
            self._private_key_pem
        )

        retriever = get_static_retriever_class({
            'client-app/key01': self._public_key_pem
        })
        self.config = {
            'ASAP_VALID_AUDIENCE': 'server-app',
            'ASAP_VALID_ISSUERS': ('client-app',),
            'ASAP_KEY_RETRIEVER_CLASS': retriever
        }

    def get_app_with_middleware(self, config):
        return ASAPMiddleware(app, config)

    def send_request(self, url='/', config=None, token=None, application=None):
        """ returns the response of sending a request containing the given
            token sent in the Authorization header.
        """

        resp_info = {}

        def start_response(status, response_headers, exc_info=None):
            resp_info['status'] = status
            resp_info['headers'] = response_headers

        environ = {}
        if token:
            environ['HTTP_AUTHORIZATION'] = b'Bearer ' + token
        if application is None:
            application = self.get_app_with_middleware(config or self.config)
        return application(environ, start_response), resp_info, environ

    def test_request_with_valid_token_is_allowed(self):
        token = create_token(
            'client-app', 'server-app',
            'client-app/key01', self._private_key_pem
        )
        body, resp_info, environ = self.send_request(token=token)
        self.assertEqual(resp_info['status'], '200 OK')
        self.assertIn('ATL_ASAP_CLAIMS', environ)

    def test_request_with_duplicate_jti_is_rejected_as_per_setting(self):
        self.config['ASAP_CHECK_JTI_UNIQUENESS'] = True
        token = create_token(
            'client-app', 'server-app',
            'client-app/key01', self._private_key_pem
        )
        application = self.get_app_with_middleware(self.config)
        body, resp_info, environ = self.send_request(
            token=token, application=application)
        self.assertEqual(resp_info['status'], '200 OK')
        body, resp_info, environ = self.send_request(
            token=token, application=application)
        self.assertEqual(resp_info['status'], '401 Unauthorized')

    def _assert_request_with_duplicate_jti_is_accepted(self):
        token = create_token(
            'client-app', 'server-app',
            'client-app/key01', self._private_key_pem
        )
        application = self.get_app_with_middleware(self.config)
        body, resp_info, environ = self.send_request(
            token=token, application=application)
        self.assertEqual(resp_info['status'], '200 OK')
        body, resp_info, environ = self.send_request(
            token=token, application=application)
        self.assertEqual(resp_info['status'], '200 OK')

    def test_request_with_duplicate_jti_is_accepted(self):
        self._assert_request_with_duplicate_jti_is_accepted()

    def test_request_with_duplicate_jti_is_accepted_as_per_setting(self):
        self.config['ASAP_CHECK_JTI_UNIQUENESS'] = False
        self._assert_request_with_duplicate_jti_is_accepted()

    def test_request_with_invalid_audience_is_rejected(self):
        token = create_token(
            'client-app', 'invalid-audience',
            'client-app/key01', self._private_key_pem
        )
        body, resp_info, environ = self.send_request(token=token)
        self.assertEqual(resp_info['status'], '401 Unauthorized')
        self.assertNotIn('ATL_ASAP_CLAIMS', environ)

    def test_request_with_invalid_token_is_rejected(self):
        body, resp_info, environ = self.send_request(token=b'notavalidtoken')
        self.assertEqual(resp_info['status'], '401 Unauthorized')
        self.assertNotIn('ATL_ASAP_CLAIMS', environ)

    def test_request_subject_and_issue_not_matching(self):
        token = create_token(
            'client-app', 'server-app',
            'client-app/key01', self._private_key_pem,
            subject='different'
        )
        body, resp_info, environ = self.send_request(token=token)
        self.assertEqual(resp_info['status'], '401 Unauthorized')
        self.assertNotIn('ATL_ASAP_CLAIMS', environ)

    def test_request_subject_does_not_need_to_match_issuer_from_settings(self):
        self.config['ASAP_SUBJECT_SHOULD_MATCH_ISSUER'] = False
        token = create_token(
            'client-app', 'server-app',
            'client-app/key01', self._private_key_pem,
            subject='different'
        )
        body, resp_info, environ = self.send_request(token=token)
        self.assertEqual(resp_info['status'], '200 OK')
        self.assertIn('ATL_ASAP_CLAIMS', environ)



import asyncio
import urllib.parse

import aiohttp

from atlassian_jwt_auth.exceptions import PublicKeyRetrieverException
from atlassian_jwt_auth.key import (
    PEM_FILE_TYPE,
    HTTPSPublicKeyRetriever as _HTTPSPublicKeyRetriever
)


class HTTPSPublicKeyRetriever(_HTTPSPublicKeyRetriever):
    """A class for retrieving JWT public keys with aiohttp"""
    _class_session = None

    def __init__(self, base_url, *, loop=None):
        if loop is None:
            loop = asyncio.get_event_loop()
        self.loop = loop
        super().__init__(base_url)

    def _get_session(self):
        if HTTPSPublicKeyRetriever._class_session is None:
            HTTPSPublicKeyRetriever._class_session = aiohttp.ClientSession(
                loop=self.loop)
        return HTTPSPublicKeyRetriever._class_session

    def _convert_proxies_to_proxy_arg(self, url, requests_kwargs):
        """ returns a modified requests_kwargs dict that contains proxy
            information in a form that aiohttp accepts
            (it wants proxy information instead of a dict of proxies).
        """
        proxy = None
        if 'proxies' in requests_kwargs:
            scheme = urllib.parse.urlparse(url).scheme
            proxy = requests_kwargs['proxies'].get(scheme, None)
            del requests_kwargs['proxies']
            requests_kwargs['proxy'] = proxy
        return requests_kwargs

    async def _retrieve(self, url, requests_kwargs):
        requests_kwargs = self._convert_proxies_to_proxy_arg(
            url, requests_kwargs)
        try:
            resp = await self._session.get(url, headers={'accept':
                                                         PEM_FILE_TYPE},
                                           **requests_kwargs)
            resp.raise_for_status()
            self._check_content_type(url, resp.headers['content-type'])
            return await resp.text()
        except aiohttp.ClientError as e:
            status_code = getattr(e, 'code', None)
            raise PublicKeyRetrieverException(e, status_code=status_code)



class SettingsDict(dict):
    def __getattr__(self, name):
        if name not in self:
            raise AttributeError

        return self[name]

    def __setitem__(self, key, value):
        raise AttributeError('SettingsDict properties are immutable')

    def _hash_key(self):
        keys_and_values = []
        for key, value in self.items():
            if isinstance(value, set):
                value = frozenset(value)
            keys_and_values.append("%s %s" % (key, hash(value)))
        return frozenset(keys_and_values)

    def __hash__(self):
        return hash(self._hash_key())

    def __eq__(self, other):
        return hash(self) == hash(other)



import warnings


warnings.warn(
    "The atlassian_jwt_auth.contrib.django package is deprecated in 4.0.0 "
    "in favour of atlassian_jwt_auth.frameworks.django.",
    DeprecationWarning, stacklevel=2
)



from django.conf import settings as django_settings
from django.http import HttpResponse, HttpResponseForbidden

from ..common.backend import Backend


class DjangoBackend(Backend):
    def get_authorization_header(self, request=None):
        if request is None:
            raise ValueError('No request available')

        return request.META.get('HTTP_AUTHORIZATION', b'')

    def get_401_response(self, data=None, headers=None, request=None):
        if headers is None:
            headers = {}

        headers.update(self.default_headers_401)

        response = HttpResponse(content=data, status=401)
        for k, v in headers.items():
            response[k] = v

        return response

    def get_403_response(self, data=None, headers=None, request=None):
        if headers is None:
            headers = {}

        response = HttpResponseForbidden(data)
        for k, v in headers.items():
            response[k] = v

        return response

    def set_asap_claims_for_request(self, request, claims):
        request.asap_claims = claims

    @property
    def settings(self):
        settings = {}
        settings.update(self.default_settings)

        for k in settings.keys():
            value = getattr(django_settings, k, None)
            if value is None:
                continue

            settings[k] = value

        return self._process_settings(settings)



from aiohttp import BasicAuth

from atlassian_jwt_auth.auth import BaseJWTAuth


class JWTAuth(BaseJWTAuth, BasicAuth):
    """Adds a JWT bearer token to the request per the ASAP specification

    It should be aiohttp.BasicAuth subclass, so redefine its `__new__` method.
    """
    def __new__(cls, *args, **kwargs):
        return super().__new__(cls, '')

    def encode(self):
        return self._get_header_value().decode(self.encoding)


def create_jwt_auth(
        issuer, key_identifier, private_key_pem, audience, **kwargs):
    """Instantiate a JWTAuth while creating the signer inline"""
    return JWTAuth.create(
        issuer, key_identifier, private_key_pem, audience, **kwargs)



import base64
import cgi
import logging
import os
import re
from urllib.parse import unquote_plus

import cachecontrol
import cryptography.hazmat.backends
import jwt
import requests
import requests.utils
from cryptography.hazmat.primitives import serialization
from requests.exceptions import RequestException, ConnectionError

from atlassian_jwt_auth.exceptions import (KeyIdentifierException,
                                           PublicKeyRetrieverException,
                                           PrivateKeyRetrieverException)


PEM_FILE_TYPE = 'application/x-pem-file'


class KeyIdentifier(object):

    """ This class represents a key identifier """

    def __init__(self, identifier):
        self.__key_id = validate_key_identifier(identifier)

    @property
    def key_id(self):
        return self.__key_id


def validate_key_identifier(identifier):
    """ returns a validated key identifier. """
    regex = re.compile(r'^[\w.\-\+/]*$')
    _error_msg = 'Invalid key identifier %s' % identifier
    if not identifier:
        raise KeyIdentifierException(_error_msg)
    if not regex.match(identifier):
        raise KeyIdentifierException(_error_msg)
    normalised = os.path.normpath(identifier)
    if normalised != identifier:
        raise KeyIdentifierException(_error_msg)
    if normalised.startswith('/'):
        raise KeyIdentifierException(_error_msg)
    if '..' in normalised:
        raise KeyIdentifierException(_error_msg)
    return identifier


def _get_key_id_from_jwt_header(a_jwt):
    """ returns the key identifier from a jwt header. """
    header = jwt.get_unverified_header(a_jwt)
    return KeyIdentifier(header['kid'])


class BasePublicKeyRetriever(object):
    """ Base class for retrieving a public key. """

    def retrieve(self, key_identifier, **kwargs):
        raise NotImplementedError()


class HTTPSPublicKeyRetriever(BasePublicKeyRetriever):

    """ This class retrieves public key from a https location based upon the
         given key id.
    """
    # Use a static requests session, reused/shared by all instances of
    # HTTPSPublicKeyRetriever:
    _class_session = None

    def __init__(self, base_url):
        if base_url is None or not base_url.startswith('https://'):
            raise PublicKeyRetrieverException(
                'The base url must start with https://')
        if not base_url.endswith('/'):
            base_url += '/'
        self.base_url = base_url
        self._session = self._get_session()
        self._proxies = requests.utils.get_environ_proxies(self.base_url)

    def _get_session(self):
        if HTTPSPublicKeyRetriever._class_session is None:
            session = cachecontrol.CacheControl(requests.Session())
            session.trust_env = False
            HTTPSPublicKeyRetriever._class_session = session
        return HTTPSPublicKeyRetriever._class_session

    def retrieve(self, key_identifier, **requests_kwargs):
        """ returns the public key for given key_identifier. """
        if not isinstance(key_identifier, KeyIdentifier):
            key_identifier = KeyIdentifier(key_identifier)
        if self._proxies and 'proxies' not in requests_kwargs:
            requests_kwargs['proxies'] = self._proxies
        url = self.base_url + key_identifier.key_id
        try:
            return self._retrieve(url, requests_kwargs)
        except requests.RequestException as e:
            try:
                status_code = e.response.status_code
            except AttributeError:
                status_code = None
            raise PublicKeyRetrieverException(e, status_code=status_code)

    def _retrieve(self, url, requests_kwargs):
        resp = self._session.get(url, headers={'accept': PEM_FILE_TYPE},
                                 **requests_kwargs)
        resp.raise_for_status()
        self._check_content_type(url, resp.headers['content-type'])
        return resp.text

    def _check_content_type(self, url, content_type):
        media_type = cgi.parse_header(content_type)[0]

        if media_type.lower() != PEM_FILE_TYPE.lower():
            raise PublicKeyRetrieverException(
                "Invalid content-type, '%s', for url '%s' ." %
                (content_type, url))


class HTTPSMultiRepositoryPublicKeyRetriever(BasePublicKeyRetriever):
    """ This class retrieves public key from the supplied https key
        repository locations based upon key ids.
    """

    def __init__(self, key_repository_urls):
        if not isinstance(key_repository_urls, list):
            raise TypeError('keystore_urls must be a list of urls.')
        self._retrievers = self._create_retrievers(key_repository_urls)

    def _create_retrievers(self, key_repository_urls):
        return [HTTPSPublicKeyRetriever(url) for url
                in key_repository_urls]

    def handle_retrieval_exception(self, retriever, exception):
        """ Handles working with exceptions encountered during key
            retrieval.
        """
        if isinstance(exception, PublicKeyRetrieverException):
            original_exception = getattr(
                exception, 'original_exception', None)
            if isinstance(original_exception, ConnectionError):
                return
            if exception.status_code is None or exception.status_code < 500:
                raise

    def retrieve(self, key_identifier, **requests_kwargs):
        for retriever in self._retrievers:
            try:
                return retriever.retrieve(key_identifier, **requests_kwargs)
            except (RequestException, PublicKeyRetrieverException) as e:
                self.handle_retrieval_exception(retriever, e)
                logger = logging.getLogger(__name__)
                logger.warning(
                    'Unable to retrieve public key from store',
                    extra={'underlying_error': str(e),
                           'key repository': retriever.base_url})
        raise PublicKeyRetrieverException(
            'Cannot load key from key repositories')


class BasePrivateKeyRetriever(object):
    """ This is the base private key retriever class. """

    def load(self, issuer):
        """ returns the key identifier and private key pem found
            for the given issuer.
        """
        raise NotImplementedError('Not implemented.')


class DataUriPrivateKeyRetriever(BasePrivateKeyRetriever):
    """ This class can be used to retrieve the key identifier and
        private key from the supplied data uri.
    """

    def __init__(self, data_uri):
        self._data_uri = data_uri

    def load(self, issuer):
        if not self._data_uri.startswith('data:application/pkcs8;kid='):
            raise PrivateKeyRetrieverException('Unrecognised data uri format.')
        splitted = self._data_uri.split(';')
        key_identifier = KeyIdentifier(unquote_plus(
            splitted[1][len('kid='):]))
        key_data = base64.b64decode(splitted[-1].split(',')[-1])
        key = serialization.load_der_private_key(
            key_data,
            password=None,
            backend=cryptography.hazmat.backends.default_backend())
        private_key_pem = key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm=serialization.NoEncryption()
        )
        return key_identifier, private_key_pem.decode('utf-8')


class StaticPrivateKeyRetriever(BasePrivateKeyRetriever):
    """ This class simply returns the key_identifier and private_key_pem
        initially provided to it in calls to load.
    """

    def __init__(self, key_identifier, private_key_pem):
        if not isinstance(key_identifier, KeyIdentifier):
            key_identifier = KeyIdentifier(key_identifier)

        self.key_identifier = key_identifier
        self.private_key_pem = private_key_pem

    def load(self, issuer):
        return self.key_identifier, self.private_key_pem


class FilePrivateKeyRetriever(BasePrivateKeyRetriever):
    """ This class can be used to retrieve the latest key identifier and
        private key for a given issuer found under its private key
        repository path.
    """

    def __init__(self, private_key_repository_path):
        self.private_key_repository = FilePrivateKeyRepository(
            private_key_repository_path)

    def load(self, issuer):
        key_identifier = self._find_last_key_id(issuer)
        private_key_pem = self.private_key_repository.load_key(key_identifier)
        return key_identifier, private_key_pem

    def _find_last_key_id(self, issuer):
        key_identifiers = list(
            self.private_key_repository.find_valid_key_ids(issuer))

        if key_identifiers:
            return key_identifiers[-1]
        else:
            raise IOError('Issuer has no valid keys: %s' % issuer)


class FilePrivateKeyRepository(object):
    """ This class represents a file backed private key repository. """

    def __init__(self, path):
        self.path = path

    def find_valid_key_ids(self, issuer):
        issuer_directory = os.path.join(self.path, issuer)
        for filename in sorted(os.listdir(issuer_directory)):
            if filename.endswith('.pem'):
                yield KeyIdentifier('%s/%s' % (issuer, filename))

    def load_key(self, key_identifier):
        key_filename = os.path.join(self.path, key_identifier.key_id)
        with open(key_filename, 'rb') as f:
            return f.read().decode('utf-8')



from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives import serialization

import atlassian_jwt_auth


def get_new_rsa_private_key_in_pem_format():
    """ returns a new rsa key in pem format. """
    private_key = rsa.generate_private_key(
        key_size=2048, backend=default_backend(), public_exponent=65537)
    return private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=serialization.NoEncryption()
    )


def get_public_key_pem_for_private_key_pem(private_key_pem):
    private_key = serialization.load_pem_private_key(
        private_key_pem,
        password=None,
        backend=default_backend()
    )
    public_key = private_key.public_key()
    return public_key.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo
    )


def get_example_jwt_auth_signer(**kwargs):
    """ returns an example jwt_auth_signer instance. """
    issuer = kwargs.get('issuer', 'egissuer')
    key_id = kwargs.get('key_id', '%s/a' % issuer)
    key = kwargs.get(
        'private_key_pem', get_new_rsa_private_key_in_pem_format())
    algorithm = kwargs.get('algorithm', 'RS256')
    return atlassian_jwt_auth.create_signer(
        issuer, key_id, key, algorithm=algorithm)


def create_token(issuer, audience, key_id, private_key, subject=None):
    """" returns a token based upon the supplied parameters. """
    signer = atlassian_jwt_auth.create_signer(
        issuer, key_id, private_key, subject=subject)
    return signer.generate_jwt(audience)


class BaseJWTAlgorithmTestMixin(object):

    """ A mixin class to make testing different support for different
        jwt algorithms easier.
    """

    def get_new_private_key_in_pem_format(self):
        """ returns a new private key in pem format. """
        raise NotImplementedError("not implemented.")


class RS256KeyTestMixin(object):

    """ Private rs256 test mixin. """

    @property
    def algorithm(self):
        return 'RS256'

    def get_new_private_key_in_pem_format(self):
        return get_new_rsa_private_key_in_pem_format()


class ES256KeyTestMixin(object):

    """ Private es256 test mixin. """

    @property
    def algorithm(self):
        return 'ES256'

    def get_new_private_key_in_pem_format(self):
        private_key = ec.generate_private_key(
            ec.SECP256R1(), default_backend())
        return private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm=serialization.NoEncryption()
        )



from atlassian_jwt_auth.frameworks.flask.decorators import with_asap


def requires_asap(f, issuers=None, subject_should_match_issuer=None):
    """
    Wrapper for Flask endpoints to make them require asap authentication to
    access.
    """

    return with_asap(func=f,
                     required=True,
                     issuers=issuers,
                     subject_should_match_issuer=subject_should_match_issuer)



import atlassian_jwt_auth


def get_static_retriever_class(keys):

    class StaticPublicKeyRetriever(object):
        """ Retrieves a key from a static list of public keys
        (for use in tests only) """

        def __init__(self, *args, **kwargs):
            self.keys = keys

        def retrieve(self, key_identifier, **requests_kwargs):
            return self.keys[key_identifier.key_id]

    return StaticPublicKeyRetriever


def static_verifier(keys):
    return atlassian_jwt_auth.JWTAuthVerifier(
        get_static_retriever_class(keys)()
    )



import datetime
import unittest
from unittest import mock

import jwt
import jwt.algorithms
import jwt.exceptions

import atlassian_jwt_auth
import atlassian_jwt_auth.exceptions
import atlassian_jwt_auth.key
import atlassian_jwt_auth.signer
from atlassian_jwt_auth.tests import utils


class NoneAlgorithmJwtAuthSigner(atlassian_jwt_auth.signer.JWTAuthSigner):
    """ A JWTAuthSigner that generates JWTs using the none algorithm
        and supports specifying arbitrary alg jwt header values.
    """

    def generate_jwt(self, audience, **kwargs):
        alg_header = kwargs.get('alg_header', 'none')
        key_identifier, private_key_pem = self.private_key_retriever.load(
            self.issuer)
        return jwt.encode(self._generate_claims(audience, **kwargs),
                          algorithm=None,
                          key=None,
                          headers={'kid': key_identifier.key_id,
                                   'alg': alg_header})


class BaseJWTAuthVerifierTest(object):

    """ tests for the JWTAuthVerifier class. """

    def setUp(self):
        self._private_key_pem = self.get_new_private_key_in_pem_format()
        self._public_key_pem = utils.get_public_key_pem_for_private_key_pem(
            self._private_key_pem)
        self._example_aud = 'aud_x'
        self._example_issuer = 'egissuer'
        self._example_key_id = '%s/a' % self._example_issuer
        self._jwt_auth_signer = atlassian_jwt_auth.create_signer(
            self._example_issuer,
            self._example_key_id,
            self._private_key_pem.decode(),
            algorithm=self.algorithm
        )

    def _setup_mock_public_key_retriever(self, pub_key_pem):
        m_public_key_ret = mock.Mock()
        m_public_key_ret.retrieve.return_value = pub_key_pem.decode()
        return m_public_key_ret

    def _setup_jwt_auth_verifier(self, pub_key_pem, **kwargs):
        m_public_key_ret = self._setup_mock_public_key_retriever(pub_key_pem)
        return atlassian_jwt_auth.JWTAuthVerifier(m_public_key_ret, **kwargs)

    def test_verify_jwt_with_valid_jwt(self):
        """ test that verify_jwt verifies a valid jwt. """
        verifier = self._setup_jwt_auth_verifier(self._public_key_pem)
        signed_jwt = self._jwt_auth_signer.generate_jwt(
            self._example_aud)
        v_claims = verifier.verify_jwt(signed_jwt, self._example_aud)
        self.assertIsNotNone(v_claims)
        self.assertEqual(v_claims['aud'], self._example_aud)
        self.assertEqual(v_claims['iss'], self._example_issuer)

    def test_verify_jwt_with_none_algorithm(self):
        """ tests that verify_jwt does not accept jwt that use the none
            algorithm.
        """
        verifier = self._setup_jwt_auth_verifier(self._public_key_pem)
        private_key_ret = atlassian_jwt_auth.key.StaticPrivateKeyRetriever(
            self._example_key_id, self._private_key_pem.decode())
        jwt_signer = NoneAlgorithmJwtAuthSigner(
            issuer=self._example_issuer,
            private_key_retriever=private_key_ret,
        )
        for algorithm in ['none', 'None', 'nOne', 'nonE', 'NONE']:
            if algorithm != 'none':
                jwt.register_algorithm(
                    algorithm, jwt.algorithms.NoneAlgorithm())
            jwt_token = jwt_signer.generate_jwt(
                self._example_aud, alg_header=algorithm)
            if algorithm != 'none':
                jwt.unregister_algorithm(algorithm)
            jwt_headers = jwt.get_unverified_header(jwt_token)
            self.assertEqual(jwt_headers['alg'], algorithm)
            with self.assertRaises(jwt.exceptions.InvalidAlgorithmError):
                verifier.verify_jwt(jwt_token, self._example_aud)

    def test_verify_jwt_with_key_identifier_not_starting_with_issuer(self):
        """ tests that verify_jwt rejects a jwt if the key identifier does
            not start with the claimed issuer.
        """
        verifier = self._setup_jwt_auth_verifier(self._public_key_pem)
        signer = atlassian_jwt_auth.create_signer(
            'issuer', 'issuerx', self._private_key_pem.decode(),
            algorithm=self.algorithm,
        )
        a_jwt = signer.generate_jwt(self._example_aud)
        with self.assertRaisesRegex(ValueError, 'Issuer does not own'):
            verifier.verify_jwt(a_jwt, self._example_aud)

    @mock.patch('atlassian_jwt_auth.verifier.jwt.decode')
    def test_verify_jwt_with_non_matching_sub_and_iss(self, m_j_decode):
        """ tests that verify_jwt rejects a jwt if the claims
            contains a subject which does not match the issuer.
        """
        expected_msg = 'Issuer does not match the subject'
        m_j_decode.return_value = {
            'iss': self._example_issuer,
            'sub': self._example_issuer[::-1]
        }
        a_jwt = self._jwt_auth_signer.generate_jwt(self._example_aud)
        verifier = self._setup_jwt_auth_verifier(self._public_key_pem)
        for exception in [
            ValueError,
            atlassian_jwt_auth.exceptions.SubjectDoesNotMatchIssuerException,
        ]:
            with self.assertRaisesRegex(exception, expected_msg):
                verifier.verify_jwt(a_jwt, self._example_aud)

    @mock.patch('atlassian_jwt_auth.verifier.jwt.decode')
    def test_verify_jwt_with_jwt_lasting_gt_max_time(self, m_j_decode):
        """ tests that verify_jwt rejects a jwt if the claims
            period of validity is greater than the allowed maximum.
        """
        expected_msg = 'exceeds the maximum'
        claims = self._jwt_auth_signer._generate_claims(self._example_aud)
        claims['iat'] = claims['exp'] - datetime.timedelta(minutes=61)
        for key in ['iat', 'exp']:
            claims[key] = claims[key].strftime('%s')
        m_j_decode.return_value = claims
        a_jwt = self._jwt_auth_signer.generate_jwt(self._example_aud)
        verifier = self._setup_jwt_auth_verifier(self._public_key_pem)
        with self.assertRaisesRegex(ValueError, expected_msg):
            verifier.verify_jwt(a_jwt, self._example_aud)

    def test_verify_jwt_with_jwt_with_already_seen_jti(self):
        """ tests that verify_jwt rejects a jwt if the jti
            has already been seen.
        """
        verifier = self._setup_jwt_auth_verifier(
            self._public_key_pem, check_jti_uniqueness=True)
        a_jwt = self._jwt_auth_signer.generate_jwt(
            self._example_aud)
        self.assertIsNotNone(verifier.verify_jwt(
            a_jwt,
            self._example_aud))
        for exception in [
                ValueError,
                atlassian_jwt_auth.exceptions.JtiUniquenessException]:
            with self.assertRaisesRegex(exception, 'has already been used'):
                verifier.verify_jwt(a_jwt, self._example_aud)

    def assert_jwt_accepted_more_than_once(self, verifier, a_jwt):
        """ asserts that the given jwt is accepted more than once. """
        for i in range(0, 3):
            self.assertIsNotNone(
                verifier.verify_jwt(a_jwt, self._example_aud))

    def test_verify_jwt_with_already_seen_jti_with_uniqueness_disabled(self):
        """ tests that verify_jwt accepts a jwt if the jti
            has already been seen and the verifier has been set
            to not check the uniqueness of jti.
        """
        verifier = self._setup_jwt_auth_verifier(
            self._public_key_pem, check_jti_uniqueness=False)
        a_jwt = self._jwt_auth_signer.generate_jwt(self._example_aud)
        self.assert_jwt_accepted_more_than_once(verifier, a_jwt)

    def test_verify_jwt_with_already_seen_jti_default(self):
        """ tests that verify_jwt by default accepts a jwt if the jti
            has already been seen.
        """
        verifier = self._setup_jwt_auth_verifier(
            self._public_key_pem)
        a_jwt = self._jwt_auth_signer.generate_jwt(self._example_aud)
        self.assert_jwt_accepted_more_than_once(verifier, a_jwt)

    def test_verify_jwt_subject_should_match_issuer(self):
        verifier = self._setup_jwt_auth_verifier(
            self._public_key_pem, subject_should_match_issuer=True)
        a_jwt = self._jwt_auth_signer.generate_jwt(
            self._example_aud,
            additional_claims={'sub': 'not-' + self._example_issuer})
        with self.assertRaisesRegex(ValueError,
                                    'Issuer does not match the subject.'):
            verifier.verify_jwt(a_jwt, self._example_aud)

    def test_verify_jwt_subject_does_not_need_to_match_issuer(self):
        verifier = self._setup_jwt_auth_verifier(
            self._public_key_pem, subject_should_match_issuer=False)
        a_jwt = self._jwt_auth_signer.generate_jwt(
            self._example_aud,
            additional_claims={'sub': 'not-' + self._example_issuer})
        self.assertIsNotNone(verifier.verify_jwt(a_jwt, self._example_aud))

    @mock.patch('atlassian_jwt_auth.verifier.jwt.decode')
    def test_verify_jwt_with_missing_aud_claim(self, m_j_decode):
        """ tests that verify_jwt rejects jwt that do not have an aud
            claim.
        """
        expected_msg = ('Claims validity, the aud claim must be provided and '
                        'cannot be empty.')
        claims = self._jwt_auth_signer._generate_claims(self._example_aud)
        del claims['aud']
        m_j_decode.return_value = claims
        a_jwt = self._jwt_auth_signer.generate_jwt(self._example_aud)
        verifier = self._setup_jwt_auth_verifier(self._public_key_pem)
        with self.assertRaisesRegex(KeyError, expected_msg):
            verifier.verify_jwt(a_jwt, self._example_aud)

    def test_verify_jwt_with_none_aud(self):
        """ tests that verify_jwt rejects jwt that have a None aud claim. """
        verifier = self._setup_jwt_auth_verifier(self._public_key_pem)
        a_jwt = self._jwt_auth_signer.generate_jwt(
            self._example_aud,
            additional_claims={'aud': None})
        exceptions = (jwt.exceptions.InvalidAudienceError,
                      jwt.exceptions.InvalidTokenError)
        with self.assertRaises(exceptions) as cm:
            verifier.verify_jwt(a_jwt, self._example_aud)
        if not isinstance(cm.exception, jwt.exceptions.InvalidAudienceError):
            self.assertIn('aud', str(cm.exception))

    def test_verify_jwt_with_non_matching_aud(self):
        """ tests that verify_jwt rejects a jwt if the aud claim does not
            match the given & expected audience.
        """
        verifier = self._setup_jwt_auth_verifier(self._public_key_pem)
        a_jwt = self._jwt_auth_signer.generate_jwt(
            self._example_aud,
            additional_claims={'aud': self._example_aud + '-different'})
        with self.assertRaises(jwt.exceptions.InvalidAudienceError):
            verifier.verify_jwt(a_jwt, self._example_aud)


class JWTAuthVerifierRS256Test(
        BaseJWTAuthVerifierTest,
        utils.RS256KeyTestMixin,
        unittest.TestCase):
    pass


class JWTAuthVerifierES256Test(
        BaseJWTAuthVerifierTest,
        utils.ES256KeyTestMixin,
        unittest.TestCase):
    pass



def get_permitted_algorithm_names():
    """ returns permitted algorithm names. """
    return [
        'RS256',
        'RS384',
        'RS512',
        'ES256',
        'ES384',
        'ES512',
        'PS256',
        'PS384',
        'PS512'
    ]



from ..common.asap import _process_asap_token
from .backend import DjangoBackend


def asap_middleware(get_response):
    """Middleware to enable ASAP for all requests"""
    backend = DjangoBackend()
    settings = backend.settings
    _verifier = backend.get_verifier(settings=settings)

    def middleware(request):
        error_response = _process_asap_token(request, backend, settings,
                                             verifier=_verifier)
        if error_response is not None:
            return error_response

        return get_response(request)

    return middleware


class OldStyleASAPMiddleware(object):
    """Middleware to enable ASAP for all requests (for legacy applications
    using MIDDLEWARE_CLASSES)"""

    def __init__(self):
        self.backend = DjangoBackend()
        self.settings = self.backend.settings
        self._verifier = self.backend.get_verifier(settings=self.settings)

    def process_request(self, request):
        error_response = _process_asap_token(
            request, self.backend, self.settings, verifier=self._verifier
        )
        if error_response is not None:
            return error_response



from collections import OrderedDict
from functools import lru_cache

import jwt
import jwt.api_jwt
from cryptography.hazmat.primitives.asymmetric.ec import (
    EllipticCurvePublicKey
)
from cryptography.hazmat.primitives.asymmetric.rsa import (
    RSAPublicKey
)
from jwt.exceptions import InvalidAlgorithmError

from atlassian_jwt_auth import algorithms
from atlassian_jwt_auth import key
from atlassian_jwt_auth import exceptions


@lru_cache(maxsize=10)
def _load_public_key(algorithms, public_key, algorithm):
    """ Returns a public key object instance given the public key and
        algorithm.

        This has been extracted out of JWTAuthVerifier to avoid possible memory
        leaks via retained instance references.
    """
    if isinstance(public_key, (RSAPublicKey, EllipticCurvePublicKey)):
        return public_key
    if algorithm not in algorithms:
        raise InvalidAlgorithmError(
            'The specified alg value is not allowed')
    py_jws = jwt.api_jws.PyJWS(algorithms=algorithms)
    alg_obj = py_jws._algorithms[algorithm]
    return alg_obj.prepare_key(public_key)


class JWTAuthVerifier(object):

    """ This class can be used to verify a JWT. """

    def __init__(self, public_key_retriever, **kwargs):
        self.public_key_retriever = public_key_retriever
        self.algorithms = algorithms.get_permitted_algorithm_names()
        self._seen_jti = OrderedDict()
        self._subject_should_match_issuer = kwargs.get(
            'subject_should_match_issuer', True)
        self._check_jti_uniqueness = kwargs.get(
            'check_jti_uniqueness', False)

    def verify_jwt(self, a_jwt, audience, leeway=0, **requests_kwargs):
        """Verify if the token is correct

        Returns:
             dict: the claims of the given jwt if verification is successful.

        Raises:
            ValueError: if verification failed.
        """
        key_identifier = key._get_key_id_from_jwt_header(a_jwt)
        public_key = self._retrieve_pub_key(key_identifier, requests_kwargs)

        alg = jwt.get_unverified_header(a_jwt).get('alg', None)
        public_key_obj = self._load_public_key(public_key, alg)
        return self._decode_jwt(
            a_jwt, key_identifier, public_key_obj,
            audience=audience, leeway=leeway)

    def _retrieve_pub_key(self, key_identifier, requests_kwargs):
        return self.public_key_retriever.retrieve(
            key_identifier, **requests_kwargs)

    def _load_public_key(self, public_key, algorithm):
        """ Returns a public key object instance given the public key and
            algorithm.
        """
        return _load_public_key(tuple(self.algorithms), public_key, algorithm)

    def _decode_jwt(self, a_jwt, key_identifier, jwt_key,
                    audience=None, leeway=0):
        """Decode JWT and check if it's valid"""
        options = {
            'verify_signature': True,
            'require': ['exp', 'iat'],
            'require_exp': True,
            'require_iat': True,
        }

        claims = jwt.decode(
            a_jwt,
            key=jwt_key,
            algorithms=self.algorithms,
            options=options,
            audience=audience,
            leeway=leeway)

        if (not key_identifier.key_id.startswith('%s/' % claims['iss']) and
                key_identifier.key_id != claims['iss']):
            raise ValueError('Issuer does not own the supplied public key')

        if self._subject_should_match_issuer and (
                claims.get('sub') and claims['iss'] != claims['sub']):
            raise exceptions.SubjectDoesNotMatchIssuerException(
                'Issuer does not match the subject.')

        _aud = claims.get('aud', None)
        if _aud is None:
            _msg = ("Claims validity, the aud claim must be provided and "
                    "cannot be empty.")
            raise KeyError(_msg)
        _exp = int(claims['exp'])
        _iat = int(claims['iat'])
        if _exp - _iat > 3600:
            _msg = ("Claims validity, '%s', exceeds the maximum 1 hour." %
                    (_exp - _iat))
            raise ValueError(_msg)
        _jti = claims['jti']
        if self._check_jti_uniqueness:
            self._check_jti(_jti)
        return claims

    def _check_jti(self, jti):
        """Checks that the given jti has not been already been used."""
        if jti in self._seen_jti:
            raise exceptions.JtiUniquenessException(
                "The jti, '%s', has already been used." % jti)
        self._seen_jti[jti] = None
        while len(self._seen_jti) > 1000:
            self._seen_jti.popitem(last=False)
